

(*
 * makeChange : int list -> int -> int list option
 * REQUIRES: all elements of coins > 0, target >= 0
 * ENSURES: makeChange coins target ==> SOME lst 
 *  where lst is a list of elements from coins 
 *  (allowing duplicates) summing up to target, 
 *  or ==> NONE if no such list exists 
 *)
fun makeChange (coins:int list) (target:int) : int list option = 
  if target = 0 
  then SOME [] 
  else 
    (case coins of 
        [] => NONE 
      | c::cs => 
        (case makeChange cs target of
            SOME lst => SOME lst
          | NONE => 
            if c > target
            then NONE
            else
              (case makeChange coins (target - c) of 
                  SOME lst => SOME (c::lst)
                | NONE => NONE
              )
        )
    )



(*
 * makeChangeCPSHelper : 
 *  int list 
 *  -> int 
 *  -> (int list -> int list option) 
 *  -> (unit -> int list option)
 *  -> int list option
 * REQUIRES: all elements of coins > 0, target >= 0
 * ENSURES: makeChange coint target s f ==> s lst 
 *  where lst is a list of elements from coins 
 *  (allowing duplicates) summing up to target, 
 *  or ==> f () if no such list exists 
 *)
fun makeChangeCPSHelper 
  (coins:int list) 
  (target:int) 
  (success:int list -> int list option) 
  (failure:unit -> int list option)
: int list option =
  if target = 0
  then success [] 
  else 
    (case coins of 
        [] => failure ()
      | c::cs =>
        makeChangeCPSHelper cs target success
          (if c > target 
          then failure 
          else fn () => 
            makeChangeCPSHelper coins (target - c)
              (fn lst => success (c::lst))
              failure
          )
    )

(*
 * makeChangeCPS : int list -> int -> int list option
 * REQUIRES: all elements of coins > 0, target >= 0
 * ENSURES: makeChangeCPS coins target ==> SOME lst 
 *  where lst is a list of elements from coins 
 *  (allowing duplicates) summing up to target, 
 *  or ==> NONE if no such list exists 
 *)
fun makeChangeCPS (coins:int list) (target:int) : int list option =
  makeChangeCPSHelper coins target SOME (fn () => NONE)






exception Change

(*
 * makeChangeExnHelper : int list -> int -> int list
 * REQUIRES: all elements of coins > 0, target >= 0
 * ENSURES: makeChangeExnHelper coins target ==> lst
 *  where lst is a list of elements from coins 
 *  (allowing duplicates) summing up to target, 
 *  or raises Change if no such list exists 
 *)
fun makeChangeExnHelper (coins:int list) (target:int) : int list = 
  if target = 0 
  then []
  else 
    (case coins of 
        [] => raise Change
      | c::cs => makeChangeExnHelper cs target 
          handle Change => 
            if c > target 
            then raise Change
            else c :: (makeChangeExnHelper coins (target - c))
    )

(*
 * makeChangeExn : int list -> int -> int list option
 * REQUIRES: all elements of coins > 0, target >= 0
 * ENSURES: makeChangeExn coins target ==> SOME lst 
 *  where lst is a list of elements from coins 
 *  (allowing duplicates) summing up to target, 
 *  or ==> NONE if no such list exists 
 *)
fun makeChangeExn (coins:int list) (target:int) : int list option = 
  SOME (makeChangeExnHelper coins target)
  handle Change => NONE
