


(*
 * firstN : int -> 'a list -> 'a list option
 * REQUIRES: n >= 0 
 * ENSURES: firstN n x1::x2::...::xn::xs ==> SOME (x1::x2::...::xn)
 *  and otherwise NONE for shorter input lst
 *)
fun firstN (n:int) (lst:'a list) : 'a list option = 
  if n = 0 
  then SOME []
  else
    (case lst of 
        [] => NONE 
      | x::xs => 
        (case firstN (n-1) xs of 
            SOME xs' => SOME (x::xs')
          | NONE => NONE
        )
    )




(*
 * firstN : int -> 'a list -> ('a list -> 'a list option) -> 'a list option
 * REQUIRES: n >= 0 
 * ENSURES: firstN n x1::x2::...::xn::xs k ==> k (x1::x2::...::xn)
 *  and otherwise NONE for shorter input lst
 *)
fun firstNCPS (n:int) (lst:'a list) (k:'a list -> 'b) (b:'b) : 'b =
  if n = 0 
  then k []
  else 
    (case lst of 
        [] => b
      | x::xs => firstNCPS (n-1) xs (fn xs' => k (x::xs')) b
    )


(*
 * firstN' : int -> 'a list -> 'a list option
 * REQUIRES: n >= 0 
 * ENSURES: firstN' n x1::x2::...::xn::xs ==> SOME (x1::x2::...::xn)
 *  and otherwise NONE for shorter input lst
 *)
fun firstN' (n:int) (lst:'a list) : 'a list option = 
  firstNCPS n lst SOME NONE




(*
 * firstNExn : int -> 'a list -> 'a list
 * REQUIRES: n >= 0 
 * ENSURES: firstNExn n x1::x2::...::xn::xs ==> x1::x2::...::xn
 *  and otherwise raises a Match exception for shorter input lst
 *)
fun firstNExn (n:int) (lst:'a list) : 'a list = 
  if n = 0 
  then []
  else 
    (case lst of 
      x::xs => x::(firstNExn (n - 1) xs)
    )


(*
 * firstNSafe : int -> 'a list -> 'a list option
 * REQUIRES: n >= 0 
 * ENSURES: firstNSafe n x1::x2::...::xn::xs ==> SOME (x1::x2::...::xn)
 *  and otherwise NONE for shorter input lst
 *)
fun firstNSafe (n:int) (lst:'a list) : 'a list option = 
  SOME (firstNExn n lst)
  handle Match => NONE